/*
 *	ldd.c
 *	list dynamic library dependencies
 *	for HP-UX 9.X
 *	see ld(1),a.out_800(4),chatr(1),dld.sl(5)
 */
static char *rcsid="$Id: ldd.c,v 1.9 1996-10-15 22:57:35+09 kabe Exp $";
static char *version="$VERSION: ldd-hp9-1.3 $";
#include <stdio.h>
#include <string.h>
#include <stdlib.h>	/*getenv*/
#include <a.out.h>
#include <shl.h>

#include <fcntl.h>	/*open*/
#include <sys/mman.h>	/*mmap*/
#include <sys/stat.h>	/*fstat*/
#include <sys/param.h>	/*MAXPATHLEN*/

#include "ldd.h"

/* local prototypes */
void	ldd(char *filename);
void	dumpShlibInfo(caddr_t dl_header_loc);
char	*basename(char	*path);
char	*libBasename(char *basename);
void	qerror(char *format,char *arg1);

int verbose = 0;
int showfilename = 1;

int main(argc,argv)
int	argc;
char	*argv[];
{
	int	c;
	extern int optind,optopt;
	char	*usage = "Usage: ldd [-v] filename ...";

	while((c=getopt(argc,argv,":v"))!=-1){
		switch(c){
		case 'v':	verbose=1; break;
		default:
			fprintf(stderr,"%s\n",usage);
			exit(1);
		}
	}
	if(optind==argc){ fprintf(stderr,"%s\n",usage); exit(1);}
	if(optind+1==argc)/*only one file*/showfilename=0;

	for( ; optind<argc; optind++){
		ldd(argv[optind]);
	}
	return 0;
}

void ldd(char *filename)
{
	int	a_out_fd;
	struct stat	a_out_stat;
	caddr_t	a_out = NULL;	/* mmap-ed target */
	struct header *header;	/* a.out header */
	char	*space_strings;	/* strings table in the header */
	SCNHDR *subspH;		/* subspace header list */
	int	i;
	char	*system;
	char	*type;

	/**************** load the target ***/
	a_out_fd = open( filename, O_RDONLY );
	if ( a_out_fd < 0 ){ qerror( "ldd: %s unreadable", filename ); return; }
	fstat( a_out_fd, &a_out_stat );
	/* avoid stupid errors for opening a directory */
	if ( ! S_ISREG( a_out_stat.st_mode )) goto not_exec;

	/** MAP_SHARED should be better, but that can't mmap
	 ** an active executable;
	 ** how big is the header?? **/
	a_out = mmap( NULL, a_out_stat.st_size, PROT_READ, MAP_FILE|MAP_PRIVATE, a_out_fd, (off_t)0 );
	if ( a_out == (caddr_t)-1 ){ qerror( "ldd: cannot mmap %s", filename ); return; }

	/**************** load a.out header ***/
	header = (struct header*)(a_out + 0);

	/**************** setup verbose header ***/
	/* these are also resolved by file(1) */

	switch( header->system_id ){
	case HP9000S800_ID:  system = "PA-RISC1.0"; break;
	case _PA_RISC1_1_ID: system = "PA-RISC1.1"; break;
	case _PA_RISC1_2_ID: system = "PA-RISC1.2"; break;
	case _PA_RISC2_0_ID: system = "PA-RISC2.0"; break;
	default:
		goto not_exec;
	}
	switch( header->a_magic ){
	case EXEC_MAGIC:   type = "normal executable";	break;
	case SHARE_MAGIC:  type = "shared executable";	break;
	case DEMAND_MAGIC: type = "demand-load executable"; break;
	case DL_MAGIC:     type = "dynamic load library"; break;
	case SHL_MAGIC:    type = "shared library";	break;
	default:
		goto not_exec;
	}
	/*printf("version_id = %d\n",header.version_id);*/


	/********** read space_strings, the strings used in a.out header **/
	space_strings = (char *)(a_out + header->space_strings_location);
	
	/********** read subspace headers (see a.out_800(4))**/
	subspH = (SCNHDR*)(a_out + header->subspace_location);

	/********** scan through subspace headers for $SHLIB_INFO$ **/
	for( i=0; i < header->subspace_total; i++ ){
		if ( !strcmp( space_strings + subspH[i].STR_INDEX, _SHLIB_INFO)){
			/************* $SHLIB_INFO$ found **/
			/* usually there's only one $SHLIB_INFO$ */

			if( subspH[i].initialization_length == 0L){
				/** initalize length of $SHLIB_INFO$ subspace was zero; thus no dynamic link information */
				printf("%s: ",filename);
				(verbose) && printf("%s %s ",system,type);
				printf( "statically linked\n" );
			}else{
				(showfilename||verbose) && printf("%s: ",filename);
				(verbose) && printf("%s %s ",system,type);
				(showfilename||verbose) && printf("\n");
				dumpShlibInfo( a_out + subspH[i].file_loc_init_value);
			}
		}
	}

	a_out && munmap( a_out, a_out_stat.st_size );
	close( a_out_fd );
	return;

not_exec:
	fprintf( stderr, "ldd: %s: not an executable\n", filename);
	a_out && munmap( a_out, a_out_stat.st_size );
	close( a_out_fd );
	return;
}

void dumpShlibInfo( caddr_t dl_header_loc )
{
	struct dl_header *dl_header;
	struct shlib_list_entry *shlib_entries;
	char	*dl_string_table;
	char	dlPath[2048];
	int	i;

	/************* load dl_header **/
	/* (this section seems to be not documented in the man pages) */

	dl_header = (struct dl_header*)(dl_header_loc);

	switch ( dl_header->hdr_version ){
	  case DL_HDR_VERSION_ID:
	  case DL_HDR_VERSION_ID2: /*HPUX10*/
		break;
	  default:
		printf("dl_header.hdr_version <%d> != DL_HDR_VERSION_ID\n", dl_header->hdr_version );
		return;

	}
	/*printf("\t\thighest highwater mark = %d\n",dl_header.highwater_mark);*/

	/************* load shlib_list_entry-s **/
	shlib_entries = (struct shlib_list_entry *)( dl_header_loc + dl_header->shlib_list_loc );
	/************* load string_table used in dl_header **/
	dl_string_table = (char*)( dl_header_loc + dl_header->string_table_loc );

	/************* build library search path from ***/
	/*             embedded path and SHLIB_PATH to dlPath; see ld(1) **/
	dlPath[0]='\0';
	if ( dl_header->flags&SHLIB_PATH_ENABLE && dl_header->flags&SHLIB_PATH_FIRST ){
		strcat(dlPath,getenv("SHLIB_PATH"));
	}
	if ( dl_header->flags&EMBED_PATH_ENABLE ){
		strcat( dlPath,":");
		strcat( dlPath, dl_string_table + dl_header->embedded_path );
	}
	if ( dl_header->flags&SHLIB_PATH_ENABLE && !(dl_header->flags&SHLIB_PATH_FIRST) ){
		strcat( dlPath,":");
		strcat( dlPath, getenv("SHLIB_PATH"));
	}
	/* dlPath ~= ":/lib/pa1.1:/usr/lib/pa1.1:/lib:/usr/lib" */

	if(verbose){
		printf("\tembedded search path\t= \"%s\" (%s)\n",
			(dl_header->embedded_path)?dl_string_table+dl_header->embedded_path:"",
			(dl_header->flags&EMBED_PATH_ENABLE)?"enabled":"disabled");
		printf("\tSHLIB_PATH order\t= %s\n",(dl_header->flags&SHLIB_PATH_ENABLE)?((dl_header->flags&SHLIB_PATH_FIRST)?"first":"last"):"disabled");
		printf("\tresulting search path\t= \"%s\"\n",dlPath);
	}

	/************* scan through shlib_entries **/
	/** usually first entry(i=0) is a.out itself, so start from i=1 */
	for ( i=1 ; i < dl_header->shlib_list_count; i++){
		char	*fullLibname = dl_string_table + shlib_entries[i].shlib_name;
		char	searchPath[MAXPATHLEN];

		if ( shlib_entries[i].dash_l_reference ){
	/** search list is only searched on "-lx" and "-l: x" linking; see ld(1) **/
			char 	*dpath;
			char	dlPathCopy[2048];
			int	found = 0;

		/** scan through dlPath **/
			strcpy(dlPathCopy,dlPath);
			for( dpath=strtok(dlPathCopy,":"); dpath; dpath=strtok(NULL,":")){
				/* null path "::" removed by strtok */
				sprintf(searchPath,"%s/%s",dpath,basename(fullLibname));
				if(0==access(searchPath,R_OK)){
					found = 1;
					break;
				}
			}
			if(!found){
		/*not in search path; test embedded full name*/
				if( 0==access(fullLibname,R_OK) ){
					strcpy( searchPath, fullLibname );
					found = 1;
				}
			}
		/*** output result ***/
			printf("\t-l%s ", libBasename(basename(fullLibname)));
			verbose && printf("(%s) ",fullLibname);
			printf("=> ");
			if(!found){
				printf("(not found)");
			}else{
				printf("%s", searchPath );
				/** check if the shared-library is executable;
				 ** see dld.sl(5) **/
				if ( access( searchPath, X_OK ) < 0 )
					printf(" (not executable?)");
			}
			printf("\n");

		}else{/* not -l */

		/** libraries linked with full-path are not searched **/
		/** only directly tested (see ld(1))**/

			printf( "\t%s", fullLibname );
			if ( access(fullLibname, R_OK) ){
				printf( " (not found)" );
			}else
			if ( access(fullLibname, X_OK) ){
				printf( " (not executable?)" );
			}
			printf( "\n" );
		}/*-l*/

	}/*next i*/
}

char	*basename(char	*path)
{
	char	*ret = strrchr(path,'/');
	if(ret==NULL)	return path;
	else	return ret+1;	/*skip '/'*/
}

char	*libBasename(char	*basename)
{
	static	char	libBase[MAXPATHLEN];
	char	*p;
	strcpy(libBase,basename+strlen("lib"));
	if(NULL!=(p=strchr(libBase,'.')))	*p='\0'; /* libxxx.sl */
	return libBase;
}

void qerror(char *format,char *arg1)
{
	char	errbuf[1024];
	sprintf(errbuf,format,arg1);
	perror(errbuf);
}

